#include <gdk/gdkevents.h>
#include <stdio.h>
#include <string.h>
#include "common.h"
#include "conf.h"

gchar *euc2utf8(const gchar *str) {
  GError *error = NULL;
  gchar *result;
  
  result = g_convert (str, -1, "UTF-8", "EUC-JP", NULL, NULL, &error);
  if (!result)  {
    g_warning ("Error converting text from EUC-JP to UTF-8: %s\n", error->message);
    g_error_free (error);
  }
  
  return result;
}

gchar *utf82euc(const gchar *str) {
  GError *error = NULL;
  gchar *result;
  
  result = g_convert (str, -1, "EUC-JP", "UTF-8", NULL, NULL, &error);
  if (!result)  {
    g_warning ("Error converting text from UTF-8 to EUC-JP: %s\n", error->message);
    g_error_free (error);
  }
  
  return result;
}

gboolean isKanaChar(gunichar c) {
  if ((c >= 12448) && (c <= 12543)) return TRUE; // Katakana 
  if ((c >= 12352) && (c <= 12447)) return TRUE; // Hiragana
  return FALSE;
}
gboolean isKatakanaChar(gunichar c) {
  if ((c >= 12448) && (c <= 12543)) return TRUE; // Katakana 
  return FALSE;
}
gboolean isHiraganaChar(gunichar c) {
  if ((c >= 12352) && (c <= 12447)) return TRUE; // Hiragana
  return FALSE;
}
gboolean isKanjiChar(gunichar c) {
  if ((c >= 19968) && (c <= 40959)) return TRUE; // Kanji
  return FALSE;
}

/* Convert Hiragana -> Katakana.*/
gchar *hira2kata(gchar *hirastr) {
  gchar *hiraptr;
  gchar *kata = g_new0(gchar, strlen(hirastr) + 6);
  gchar *kataptr = kata;
  int length;

  hiraptr = hirastr;
  while (*hiraptr != 0) {
    if (isHiraganaChar(g_utf8_get_char(hiraptr)) == TRUE) {
      g_unichar_to_utf8(g_utf8_get_char(hiraptr) + 96, kataptr);
    }
    else {
      length = g_utf8_next_char(hiraptr) - hiraptr;
      strncat(kataptr, hiraptr, length);
      kataptr[length + 1] = 0;
    }
    kataptr = g_utf8_next_char(kataptr);
    hiraptr = g_utf8_next_char(hiraptr);
    if (hiraptr == NULL) break;
  } 
  return kata;
}

/* Convert fullsize kana to half width*/
gchar *full2half(gchar *instr) {
  gchar *inptr = instr;
  gchar *out = g_new0(gchar, strlen(instr));
  gchar *outptr = out;
  int length;

  printf("instr:%s\n", utf82euc(instr));

  while (*inptr != 0) {
    if (isKanaChar(g_utf8_get_char(inptr)) == TRUE) {
      printf("hex:%x\n", (int)g_utf8_get_char(inptr));
      g_unichar_to_utf8(g_utf8_get_char(inptr) - 1, outptr);
    }
    else {
      length = g_utf8_next_char(inptr) - inptr;
      strncat(outptr, inptr, length);
      outptr[length + 1] = 0;
    }
    outptr = g_utf8_next_char(outptr);
    inptr = g_utf8_next_char(inptr);
    if (inptr == NULL) break;
  } 
  printf("outstr:%s\n", utf82euc(out));
  return out;
}

gboolean ishotkey(GdkEventKey *key, int hotkey, im_ja_cfg *cfg) {
  guint state = key->state;
  
  state &= ~GDK_LOCK_MASK; 
  state &= ~GDK_MOD2_MASK;
  state &= ~GDK_MOD5_MASK;
  
  if ((state == cfg->hotkey_states[hotkey]) &&
      (key->keyval == cfg->hotkey_values[hotkey])) return TRUE;
  
  return FALSE;

}

gboolean buffer_delchar(gchar* buf) {
  if( buf == NULL ) return FALSE;

  g_utf8_prev_char(buf + strlen(buf))[0] = 0;
  return TRUE;
}

/*
gchar *get_utf8_for_keyval(guint keyval) {
  gchar *utf8strg = g_new0(gchar, 7);
  int written = g_unichar_to_utf8(gdk_keyval_to_unicode(keyval), utf8strg);
  utf8strg[written] = 0;
  return utf8strg;
}
*/

int buffer_inschar(char* buf, int maxbuf, char *ins, int pos) {
	char *repo, *end;
	int clen = strlen(buf);
	int nlen = strlen(ins);

	if (clen+nlen >= maxbuf)	{
		nlen = maxbuf - clen - 1;
	}

	clen += nlen;

	repo = buf + clen;
	end = buf + pos + nlen;

	while (repo >= end) {
		*repo = *(repo-nlen);
		--repo;
	}

	repo = buf + pos;
	end = repo + nlen;

	while (repo < end) *repo++ = *ins++;
 
	return nlen;
}

int buffer_bkspchar(char* buf, int pos) {
	int delta;
	char *prev, *at;

	at = buf + pos;
	prev = g_utf8_find_prev_char(buf, at);

	if (prev) {
		delta = at - prev;
		do {
			*prev++ = *at++;
		}
		while (*prev);
	}

	return delta;
}

int buffer_deltchar(char* buf, int pos) {
	int delta;
	char *next, *at;

	at = buf + pos;
	next = g_utf8_find_next_char(at, 0);

	if (next) {
		delta = next - at;
		do {
			*at++ = *next++;
		}
		while (*at);
	}

	return delta;
}
