#include <string.h>
#include <gconf/gconf-client.h>
#include <gdk/gdkkeysyms.h>
#include <stdlib.h>


#include "conf.h"
#include "error.h"

#define GCONF_NAMESPACE "/system/im-ja/"

#define ENUMSIZE(s) sizeof(s)/sizeof(gchar *) 

static GConfClient *gconf_client = NULL;

static const gchar *hotkey_gconf_keys[] = { //FIXME: change this to GConfEnumStringPair
  "START_CONFIGURATOR",
  "NEXT_INPUT_MODE",
  "PREV_INPUT_MODE",
  "RAW_INPUT_MODE",
  "HIRAGANA_INPUT_MODE",
  "KATAKANA_INPUT_MODE",
  "HALFKATA_INPUT_MODE",
  "ZENKAKU_INPUT_MODE",
  "CANNA_INPUT_MODE",
  "WNN_INPUT_MODE",
	"KANJI_INPUT_MODE",
  "KANJIPAD_INPUT_MODE",
  "UNCONVERT",
  "CONVERT",
  "PREV_CANDIDATE",
  "NEXT_CANDIDATE",
  "LIST_CANDIDATES",
  "PREV_BUNSETSU",
  "NEXT_BUNSETSU",
  "EXPAND_BUNSETSU",
  "SHRINK_BUNSETSU"
};

static const gchar *preedit_gconf_keys[] = {
	"normal_foreground",
	"normal_background",
	"highlighted_foreground",
	"highlighted_background"
};

GConfEnumStringPair status_win_labels[] = {
	{IM_JA_RAW_INPUT, "raw"},
	{IM_JA_HIRAGANA_INPUT, "hiragana"},
	{IM_JA_KATAKANA_INPUT, "katakana"},
	{IM_JA_HALFKATA_INPUT, "halfkata"},
	{IM_JA_ZENKAKU_INPUT, "zenkaku"},
	{IM_JA_KANJI_INPUT, "kanji"},
	{IM_JA_KANJIPAD_INPUT, "kanjipad"},
	{-1, NULL}
};

int im_ja_init_conf_handler() {
  GError *error;

  if (gconf_client != NULL) return TRUE;
  if (gconf_init(0, NULL, &error) == FALSE) {
    im_ja_print_error("GConf init failed: %s\n", error->message);
    return FALSE;
  }
  if (gconf_client == NULL) gconf_client = gconf_client_get_default();
  return TRUE;
}

int im_ja_finalize_conf_handler(im_ja_cfg *cfg) {
  im_ja_save_conf(cfg);
  return TRUE;
}


gchar *get_string_from_color(GdkColor *color) {
	gchar tmp[8];
	gchar *colorstrg = g_new0(gchar, 8);
	
	colorstrg[0] = '#';
	g_snprintf(tmp, 8, "%02X", color->red);
	colorstrg[1] = tmp[0];
	colorstrg[2] = tmp[1];

	g_snprintf(tmp, 8, "%02X", color->green);
	colorstrg[3] = tmp[0];
	colorstrg[4] = tmp[1];

	g_snprintf(tmp, 8, "%02X", color->blue);
	colorstrg[5] = tmp[0];
	colorstrg[6] = tmp[1];
	colorstrg[7] = 0;
	
	return colorstrg; 
}


static void get_default_colors(im_ja_cfg *cfg) { /* FIXME: this doesn't get the real bg and fg. Maybe gtk theme is not applied yet at this point */
  GtkStyle *style;
	//int i;
	GtkWidget *entry = gtk_entry_new();

	gtk_widget_ensure_style(entry);
	style = gtk_widget_get_style(entry);

	cfg->original_colors[0] = style->fg[0];
	cfg->original_colors[1] = style->bg[0];

	gtk_widget_destroy(entry);

	/* Ensure they are not the same */
	if ((cfg->original_colors[0].red == cfg->original_colors[1].red) 
			&& (cfg->original_colors[0].green == cfg->original_colors[1].green) 
			&& (cfg->original_colors[0].blue == cfg->original_colors[1].blue)) { 
		
		IM_JA_DEBUG("Error: Got the same colors for fg and bg!!\n");

		/* Make foreground white */
		cfg->original_colors[0].red = 0xff; 
		cfg->original_colors[0].green = 0xff; 
		cfg->original_colors[0].blue = 0xff; 

		/* Make background black */
		cfg->original_colors[1].red = 0; 
		cfg->original_colors[1].green = 0; 
		cfg->original_colors[1].blue = 0; 
	}
	else { /* If everything is ok, invert RGB values */
		cfg->original_colors[0].red = ~cfg->original_colors[0].red; 
		cfg->original_colors[0].green = ~cfg->original_colors[0].green; 
		cfg->original_colors[0].blue = ~cfg->original_colors[0].blue; 
		cfg->original_colors[1].red = ~cfg->original_colors[0].red; 
		cfg->original_colors[1].green = ~cfg->original_colors[0].green; 
		cfg->original_colors[1].blue = ~cfg->original_colors[0].blue; 
	}
}


int im_ja_load_conf(im_ja_cfg *cfg) {
  int i;
  guint keyval, keystate;
  gchar *gconf_key;
  gchar *colorstr;

  if (gconf_client == NULL) {
    IM_JA_DEBUG("gconf_client not initialized.\n");
    return FALSE;
  }
  for (i = 0; i < ENUMSIZE(hotkey_gconf_keys); i++) {
		gconf_key = g_strdup_printf(GCONF_NAMESPACE"hotkeys/%s_keyval", hotkey_gconf_keys[i]);
    keyval = gconf_client_get_int(gconf_client, gconf_key, NULL);
		//    IM_JA_DEBUG("%s: %d\n", gconf_key, keyval);
    g_free(gconf_key);
    if ((keyval != 0) && (keyval != GDK_VoidSymbol)) {
      cfg->hotkey_values[i] = keyval;
			gconf_key = g_strdup_printf(GCONF_NAMESPACE"hotkeys/%s_state", hotkey_gconf_keys[i]);
      keystate = gconf_client_get_int(gconf_client, gconf_key, NULL);
      g_free(gconf_key);
      cfg->hotkey_states[i] = keystate;
    }
  }

	for (i = 0; i < ENUMSIZE(preedit_gconf_keys); i++) {
		gconf_key = g_strdup_printf(GCONF_NAMESPACE"preedit_style/%s", preedit_gconf_keys[i]);
		colorstr = gconf_client_get_string(gconf_client, gconf_key, NULL);
		if (colorstr != NULL) {
			gdk_color_parse((const gchar*)colorstr, &cfg->preedit_colors[i]);
			//IM_JA_DEBUG("%s: %s\n", gconf_key, colorstr);
		}
		g_free(colorstr);
		g_free(gconf_key);
	}

	cfg->custom_preedit_n =	gconf_client_get_bool(gconf_client, GCONF_NAMESPACE"preedit_style/custom_normal", NULL);
	cfg->custom_preedit_hl = gconf_client_get_bool(gconf_client, GCONF_NAMESPACE"preedit_style/custom_highlight", NULL);
	get_default_colors(cfg);

	cfg->status_win_position = gconf_client_get_int(gconf_client, GCONF_NAMESPACE"status_window/position", NULL);
	cfg->status_win_coords[0] = gconf_client_get_int(gconf_client, GCONF_NAMESPACE"status_window/fixed_xpos", NULL);
	cfg->status_win_coords[1] = gconf_client_get_int(gconf_client, GCONF_NAMESPACE"status_window/fixed_ypos", NULL);

	for (i = 0; i < IM_JA_INPUT_METHODS_TOTAL; i++) {
		gconf_key = g_strdup_printf(GCONF_NAMESPACE"status_window/label_%s", gconf_enum_to_string(status_win_labels, i));
		cfg->status_win_labels[i] = gconf_client_get_string(gconf_client, gconf_key, NULL);
		//IM_JA_DEBUG("%s: %s\n", gconf_key, cfg->status_win_labels[i]);
		g_free(gconf_key);
	}

	cfg->default_conv_engine = gconf_client_get_int(gconf_client, GCONF_NAMESPACE"conv_engine/default_conv_engine", NULL);
#ifdef DISABLE_WNN
	cfg->default_conv_engine = KANJI_INPUT_CANNA;
#endif
#ifdef DISABLE_CANNA
	cfg->default_conv_engine = KANJI_INPUT_WNN;
#endif
	
	//	IM_JA_DEBUG("DEFAULT CONV ENGINE: %d\n", cfg->default_conv_engine);

	//	cfg->wnn_port =	gconf_client_get_int(gconf_client, GCONF_NAMESPACE"conv_engine/wnn_port", NULL);
	//	cfg->canna_port =	gconf_client_get_int(gconf_client, GCONF_NAMESPACE"conv_engine/canna_port", NULL);
	cfg->wnn_address = gconf_client_get_string(gconf_client, GCONF_NAMESPACE"conv_engine/wnn_address", NULL);
	cfg->wnnenvrc = gconf_client_get_string(gconf_client, GCONF_NAMESPACE"conv_engine/wnnenvrc", NULL);
	//	cfg->canna_address = gconf_client_get_string(gconf_client, GCONF_NAMESPACE"conv_engine/canna_address", NULL);

	cfg->kanjipad_enabled = gconf_client_get_bool(gconf_client, GCONF_NAMESPACE"conv_engine/kanjipad_enabled", NULL);
	cfg->kanjipad_custom_size = gconf_client_get_bool(gconf_client, GCONF_NAMESPACE"conv_engine/kanjipad_custom_size", NULL);
	cfg->kanjipad_size[0] = gconf_client_get_int(gconf_client, GCONF_NAMESPACE"conv_engine/kanjipad_width", NULL);
	cfg->kanjipad_size[1] = gconf_client_get_int(gconf_client, GCONF_NAMESPACE"conv_engine/kanjipad_height", NULL);

	cfg->startup_input_method = gconf_client_get_int(gconf_client, GCONF_NAMESPACE"other/startup_input_method", NULL);

	return TRUE;
}

int im_ja_save_conf(im_ja_cfg *cfg) {
  int i;
  gchar *gconf_key;
  gchar *colorstr;

  if (gconf_client == NULL) {
    im_ja_print_error("Couldn't save setting: gconf_client not initialized.\n");
    return FALSE;
  }
  for (i = 0; i < HOTKEYS_TOTAL; i++) {
		gconf_key = g_strdup_printf(GCONF_NAMESPACE"hotkeys/%s_keyval", hotkey_gconf_keys[i]);
    if (cfg->hotkey_values[i] != 0) {
      gconf_client_set_int(gconf_client, gconf_key, cfg->hotkey_values[i], NULL);
      //IM_JA_DEBUG("%s: %d\n", gconf_key, cfg->hotkey_values[i]);
    }
    g_free(gconf_key);
    
		gconf_key = g_strdup_printf(GCONF_NAMESPACE"hotkeys/%s_state", hotkey_gconf_keys[i]);

    if (cfg->hotkey_values[i] != 0) {
      gconf_client_set_int(gconf_client, gconf_key, cfg->hotkey_states[i], NULL);
      //IM_JA_DEBUG("%s: %d\n", gconf_key, cfg->hotkey_states[i]);
    }
    g_free(gconf_key);
  }

  for (i = 0; i < ENUMSIZE(preedit_gconf_keys); i++) {
		gconf_key = g_strdup_printf(GCONF_NAMESPACE"preedit_style/%s", preedit_gconf_keys[i]);
		//IM_JA_DEBUG("%s: %s\n", gconf_key, get_string_from_color(&cfg->preedit_colors[i]));
		colorstr = get_string_from_color(&cfg->preedit_colors[i]);
		gconf_client_set_string(gconf_client, gconf_key, colorstr, NULL);
		g_free(colorstr);
		g_free(gconf_key);
	}

	gconf_client_set_bool(gconf_client, GCONF_NAMESPACE"preedit_style/custom_normal", cfg->custom_preedit_n, NULL);
	gconf_client_set_bool(gconf_client, GCONF_NAMESPACE"preedit_style/custom_highlight", cfg->custom_preedit_hl, NULL);

	gconf_client_set_int(gconf_client, GCONF_NAMESPACE"status_window/position", cfg->status_win_position, NULL);
	gconf_client_set_int(gconf_client, GCONF_NAMESPACE"status_window/fixed_xpos", cfg->status_win_coords[0], NULL);
	gconf_client_set_int(gconf_client, GCONF_NAMESPACE"status_window/fixed_ypos", cfg->status_win_coords[1], NULL);

	
	for (i = 0; i < IM_JA_INPUT_METHODS_TOTAL; i++) {
		gconf_key = g_strdup_printf(GCONF_NAMESPACE"status_window/label_%s", gconf_enum_to_string(status_win_labels, i));
		if (cfg->status_win_labels[i] != NULL) gconf_client_set_string(gconf_client, gconf_key, cfg->status_win_labels[i], NULL);
		g_free(gconf_key);
	}
	
	gconf_client_set_int(gconf_client, GCONF_NAMESPACE"conv_engine/default_conv_engine", cfg->default_conv_engine, NULL);
	//	gconf_client_set_int(gconf_client, GCONF_NAMESPACE"conv_engine/wnn_port", cfg->wnn_port, NULL);
	//	gconf_client_set_int(gconf_client, GCONF_NAMESPACE"conv_engine/canna_port", cfg->canna_port, NULL);
	gconf_client_set_string(gconf_client, GCONF_NAMESPACE"conv_engine/wnn_address", cfg->wnn_address, NULL);
	gconf_client_set_string(gconf_client, GCONF_NAMESPACE"conv_engine/wnnenvrc", cfg->wnnenvrc, NULL);
	//	gconf_client_set_string(gconf_client, GCONF_NAMESPACE"conv_engine/canna_address", cfg->canna_address, NULL);

	gconf_client_set_bool(gconf_client, GCONF_NAMESPACE"conv_engine/kanjipad_enabled", cfg->kanjipad_enabled, NULL);
	gconf_client_set_bool(gconf_client, GCONF_NAMESPACE"conv_engine/kanjipad_custom_size", cfg->kanjipad_custom_size, NULL);
	gconf_client_set_int(gconf_client, GCONF_NAMESPACE"conv_engine/kanjipad_width", cfg->kanjipad_size[0], NULL);
	gconf_client_set_int(gconf_client, GCONF_NAMESPACE"conv_engine/kanjipad_height", cfg->kanjipad_size[1], NULL);

	gconf_client_set_int(gconf_client, GCONF_NAMESPACE"other/startup_input_method", cfg->startup_input_method, NULL);
	
	
	return TRUE;
}

