/* IM-JA Japanese Input Method Module for GTK-2.0
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Botond Botyanszki <boti@rocketmail.com>
 *
 */

#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>


#include "im-ja.h"
#include "error.h"
#include "common.h"
#include "romakana.h"
#include "hiragana-convtable.h"
#include "zenkaku-convtable.h"
#include "halfkata-convtable.h"

/* * * * * * * * * * * * * * * * */
void im_ja_kana_convert_tailing_n(IMContextJA *cn) {
	if (strlen(cn->preedit_buf) == 0) return;
	if (cn->preedit_buf[strlen(cn->preedit_buf) - 1] == 'n') {
		cn->preedit_buf[strlen(cn->preedit_buf) - 1] = 0;
		switch (cn->input_method) {
		case IM_JA_HIRAGANA_INPUT:
			g_strlcat(cn->preedit_buf, "\xE3\x82\x93", BUFFERSIZE);
			break;
		case IM_JA_KATAKANA_INPUT:
			g_strlcat(cn->preedit_buf, "\xE3\x83\xB3", BUFFERSIZE);
			break;
		case IM_JA_HALFKATA_INPUT:
			g_strlcat(cn->preedit_buf, "\xEF\xBE\x9D", BUFFERSIZE);
			break;
		}			
	}
}

void im_ja_kana_commit_converted(IMContextJA *cn, gchar *kanastr) {
	gchar *lastkana;

	/* Find any converted char */
	if (strlen(kanastr) == 0) return;
	
	lastkana = kanastr + strlen(kanastr);
	while (g_unichar_iswide(g_utf8_get_char(g_utf8_prev_char(lastkana))) == FALSE) {
		lastkana = g_utf8_prev_char(lastkana);
		if (lastkana == kanastr) break;
	}
	
	/* copy new string back to preedit buffer */
	//IM_JA_DEBUG("lastkana: %d\n", lastkana - kanastr);
	if (lastkana != kanastr + strlen(kanastr)) {
		g_strlcpy(cn->preedit_buf, lastkana, BUFFERSIZE);
		*lastkana = 0;
	}
	else *cn->preedit_buf = 0; //FIXME: this causes gnome-terminal to segfault. I have no idea why.
	//	else im_ja_set_preedit_string(cn, " ", 0, 0); //This doesn't
	// else im_ja_reset(cn);
	g_signal_emit_by_name(cn, "commit", kanastr);	
}

gboolean im_ja_kana_filter_keypress(IMContextJA *cn, GdkEventKey *key) {
  gchar *tmpstr, *tmp2str;

  if (key->type == GDK_KEY_RELEASE) {
    return FALSE;
  }
  if (key->keyval == GDK_space) {
    g_strlcat(cn->preedit_buf, " ", BUFFERSIZE); //maybe this should be zenkaku space
    g_signal_emit_by_name(cn, "preedit_changed");
    return TRUE;
  }      
  
  if (key->keyval == GDK_BackSpace) {
    if (strlen(cn->preedit_buf) > 0) {
      buffer_delchar(cn->preedit_buf);
      g_signal_emit_by_name(cn, "preedit_changed");
      return TRUE;
    }
    return FALSE;
  }
  
  if (key->keyval == GDK_Return) {
    if (strlen(cn->preedit_buf) == 0) return FALSE;
		im_ja_kana_convert_tailing_n(cn);
    if (g_utf8_validate(cn->preedit_buf, -1, NULL) != TRUE) {
      printf("Cannot commit. utf8_validate failed: %s\n", cn->preedit_buf);
    }
    else g_signal_emit_by_name(cn, "commit", cn->preedit_buf);
		im_ja_reset(cn);
    return TRUE;
  } 
  
  if (key->keyval >= GDK_exclam && key->keyval <= GDK_overline) {
		gchar utf8strg[7];
		int utf8strg_end;
		utf8strg_end = g_unichar_to_utf8(gdk_keyval_to_unicode(key->keyval), utf8strg);
		utf8strg[utf8strg_end] = 0;
		g_strlcat(cn->preedit_buf, utf8strg, BUFFERSIZE);
		switch (cn->input_method) {
		case IM_JA_HIRAGANA_INPUT:
			tmpstr = roma2kana(cn->preedit_buf, hiraganatable);
			IM_JA_DEBUG("converted \"%s\" to \"%s\"\n", cn->preedit_buf, tmpstr);
			im_ja_kana_commit_converted(cn, tmpstr);
			g_free(tmpstr);
			break;
		case IM_JA_KATAKANA_INPUT:
			tmpstr = roma2kana(cn->preedit_buf, hiraganatable);
			tmp2str = hira2kata(tmpstr);
			im_ja_kana_commit_converted(cn, tmp2str);
			g_free(tmpstr);
			g_free(tmp2str);
			break;
		case IM_JA_HALFKATA_INPUT:
			tmpstr = roma2kana(cn->preedit_buf, halfkatatable);
			im_ja_kana_commit_converted(cn, tmpstr);
			g_free(tmpstr);
			break;
		case IM_JA_ZENKAKU_INPUT:
			tmpstr = roma2kana(cn->preedit_buf, zenkakutable);
			g_strlcpy(cn->preedit_buf, tmpstr, BUFFERSIZE);
			if (g_utf8_validate(cn->preedit_buf, -1, NULL) != TRUE) {
				printf("Cannot commit. utf8_validate failed: %s\n", cn->preedit_buf);
			}
			else g_signal_emit_by_name(cn, "commit", cn->preedit_buf);
			g_free(tmpstr);
			im_ja_reset(cn);
			break;
		}
		if (g_utf8_validate(cn->preedit_buf, -1, NULL) != TRUE) {
			printf("utf8_validate failed: %s\n", cn->preedit_buf);
		}
		else g_signal_emit_by_name(cn, "preedit_changed");
		return TRUE;
  }

  return FALSE; /* Didn't handle the keypress */
}

int g_strrncmp(gchar* s1, size_t s1len, gchar* s2, size_t s2len) {

  if( s1len > s2len )	s1 += s1len-s2len;
  return strncmp(s1, s2, s2len);
}


//ascii & unicode zenkaku diff: 65248

/* Convert romaji to kana. Returns the kana string */
gchar *roma2kana(gchar *romastr, romapair kanatable[]) {
  gchar *result = NULL;
  int i = 0,s1len,s2len;

  if (romastr == NULL) return NULL;
  if (*romastr == '\0') return NULL;

  s1len = strlen(romastr);

  for (i = 0; ; i++) {
    romapair rp = kanatable[i];
    if (rp.kana[0] == '0' && rp.roma[0] == '0') break;
		s2len = strlen(rp.roma);
    // Compare end of input and sting in kana table
    if(g_strrncmp(romastr, s1len, rp.roma,s2len) == 0 ) {
      result = g_new0(gchar, BUFFERSIZE);
      strncat(result, romastr, s1len - s2len );
      g_strlcat(result, rp.kana, BUFFERSIZE);
      return result;
    }
  }
  
  return g_strdup(romastr); /* no change */
}

gchar *roma2kana_i(gchar *romastr, gint *index, romapair kanatable[])
{
	gchar *result = NULL;
	int i = 0,s1len,s2len;

	if (romastr == NULL)
		return NULL;

	result = g_new0(gchar, BUFFERSIZE);

	if (*romastr == '\0')
		return result;

	s1len = *index;

	for (i = 0; ; i++)
	{
		romapair rp = kanatable[i];
		if (rp.kana[0] == '0' && rp.roma[0] == '0')
			break;

		s2len = strlen(rp.roma);

		// Compare offset of input and string in kana table
		if(g_strrncmp(romastr, s1len, rp.roma,s2len) == 0 )
		{
			*index += strlen(rp.kana) - s2len;
			strncat(result, romastr, s1len - s2len );
			g_strlcat(result, rp.kana, BUFFERSIZE);
			g_strlcat(result, romastr+s1len, BUFFERSIZE);
			return result;
		}
	}

	// no change
	g_strlcat(result, romastr, BUFFERSIZE);
	return result;
}


gchar *roma2kana_i_lastpass(gchar *romastr, gint *index, romapair kanatable[])
{
	gchar *result = NULL;
	gchar *subst_str="\xE3\x82\x93";
	gchar *pattern="n";
	int s1len;

	if (romastr == NULL)
		return NULL;

	result = g_new0(gchar, BUFFERSIZE);

	if (*romastr == '\0')
		return result;

	s1len = *index;

	// Compare offset of input and string in kana table
	if(g_strrncmp(romastr, s1len, pattern, 1) == 0 )
	{
		*index += 3 - 1;
		strncat(result, romastr, s1len - 1 );
		g_strlcat(result, subst_str, BUFFERSIZE);
		g_strlcat(result, romastr+s1len, BUFFERSIZE);
		return result;
	}

	// no change
	g_strlcat(result, romastr, BUFFERSIZE);
	return result;
}

